//
// Copyright (c) 2004 - InfoMus Lab - DIST - University of Genova
//
// InfoMus Lab (Laboratorio di Informatica Musicale)
// DIST - University of Genova
//
// http://www.infomus.dist.unige.it
// news://infomus.dist.unige.it
// mailto:staff@infomus.dist.unige.it
//
// Developer: Gualtiero Volpe
// mailto:volpe@infomus.dist.unige.it
//
// Modified: Oct 25 2006 by Trey Harrison
// email:trey@harrisondigitalmedia.com
//
// Last modified: Oct 01 2018 by Menno Vink
// email:menno@resolume.com

#include "PluginManager.h"
#include "PluginBase.h"
#include "ParamBase.h"
#include "ParamBuffer.h"
#include "ParamOption.h"
#include "ParamRange.h"

#include <stdlib.h>
#include <memory.h>
#include <algorithm>

namespace ffgl
{
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// PluginManager constructor and destructor
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// 2019-08-24 ffgl@subpixels.com
// - ParamInfoStruct linked list replaced by params vector
// - Use initialiser list
PluginManager::PluginManager() :
	minInputs( 0U ),
	maxInputs( 0U ),
	timeSupported( true )
{
}

// 2019-07-22 ffgl@subpixels.com - use enum ParameterType
// 2019-08-24 ffgl@subpixels.com - ParamInfoStruct linked list replaced by params vector
PluginManager::~PluginManager()
{
	// params vector is owner of the params objects and is responsible to delete them
	for( auto param : params )
	{
		if( param )
			delete( param );
	}
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// PluginManager methods
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

FFUInt32 PluginManager::GetMinInputs() const
{
	return minInputs;
}

FFUInt32 PluginManager::GetMaxInputs() const
{
	return maxInputs;
}

void PluginManager::AddParam( ParamBase* param )
{
	params.push_back( param );
}

// 2019-08-24 ffgl@subpixels.com - ParamInfoStruct linked list replaced by params vector
FFUInt32 PluginManager::GetNumParams() const
{
	return static_cast< FFUInt32 >( params.size() );
}

void PluginManager::SetMinInputs( FFUInt32 minInputs )
{
	this->minInputs = minInputs;
}

void PluginManager::SetMaxInputs( FFUInt32 maxInputs )
{
	this->maxInputs = maxInputs;
}

void PluginManager::SetTimeSupported( bool supported )
{
	timeSupported = supported;
}

// 2019-07-12 ffgl@subpixels.com - Make method const
// 2019-08-24 ffgl@subpixels.com - ParamInfoStruct linked list replaced by params vector
RangeStruct PluginManager::GetParamRange( FFUInt32 index ) const
{
	// Default range in case no range is defined (paramIndex out of bounds, param is not a ParamRange)
	RangeStruct range = { 0.0f, 1.0f };

	auto param = GetParam< ParamRange >( index );
	if( param )
		range = param->GetRange();

	return range;
}

// 2019-08-24 ffgl@subpixels.com - ParamInfoStruct linked list replaced by params vector
void PluginManager::SetParamRange( FFUInt32 index, float min, float max )
{
	auto param = FindParam< ParamRange >( index );
	if( param )
		param->SetRange( { min, max } );
}

// 2019-07-12 ffgl@subpixels.com - Make method const and return const
// 2019-08-24 ffgl@subpixels.com - ParamInfoStruct linked list replaced by params vector
const char* PluginManager::GetParamName( FFUInt32 index ) const
{
	auto param = GetParam( index );
	if( !param )
		return nullptr;

	return param->GetName().c_str();
}

// 2019-07-22 ffgl@subpixels.com - use enums ParameterType
// 2019-08-24 ffgl@subpixels.com - ParamInfoStruct linked list replaced by params vector
ParameterType PluginManager::GetParamType( FFUInt32 index ) const
{
	auto param = GetParam( index );
	if( !param )
		return ParameterType::FF_FAIL;

	return param->GetType();
}

// 2019-08-24 ffgl@subpixels.com - ParamInfoStruct linked list replaced by params vector
FFResult PluginManager::GetNumParamElements( FFUInt32 index ) const
{
	auto param = GetParam( index );
	if( !param )
		return FF_FAIL;

	ParameterType paramType = param->GetType();
	if( paramType == ParameterType::FF_TYPE_BUFFER )
	{
		auto paramBuffer = dynamic_cast< const ParamBuffer* >( param );
		return (FFResult)paramBuffer->data.size();
	}
	else if( paramType == ParameterType::FF_TYPE_OPTION )
	{
		auto paramOption = dynamic_cast< const ParamOption* >( param );
		return paramOption->GetNumOptions();
	}

	return FF_FAIL;
}

// 2019-07-12 ffgl@subpixels.com - Make method const and return const
// 2019-08-24 ffgl@subpixels.com - ParamInfoStruct linked list replaced by params vector
const char* PluginManager::GetParamElementName( FFUInt32 paramIndex, FFUInt32 elemIndex ) const
{
	const char* retval = nullptr;

	auto param = GetParam( paramIndex );
	if( !param )
		return retval;

	ParameterType paramType = param->GetType();

	if( paramType == ParameterType::FF_TYPE_OPTION )
	{
		auto paramOption = dynamic_cast< const ParamOption* >( param );
		if( paramOption )
		{
			auto option = paramOption->GetOption( elemIndex );
			if( option )
				retval = option->name.c_str();
		}
	}

	return retval;
}

// 2019-07-12 ffgl@subpixels.com - use FFMixed.FloatValue member instead of pointer recasting
// 2019-08-24 ffgl@subpixels.com - ParamInfoStruct linked list replaced by params vector
FFMixed PluginManager::GetParamElementDefault( FFUInt32 paramIndex, FFUInt32 elemIndex ) const
{
	FFMixed retval = { FF_FAIL };

	auto param = GetParam( paramIndex );
	if( !param )
		return retval;

	ParameterType paramType = param->GetType();

	if( paramType == ParameterType::FF_TYPE_BUFFER )
	{
		auto paramBuffer = dynamic_cast< const ParamBuffer* >( param );
		if( paramBuffer && elemIndex < paramBuffer->data.size() )
		{
			retval.FloatValue = paramBuffer->data[ elemIndex ];
		}
	}
	else if( paramType == ParameterType::FF_TYPE_OPTION )
	{
		// Send the element INDEX to the host, not the value.
		// This way we can keep track of the index rather than element values, which may not be unique
		retval.FloatValue = static_cast< float >( elemIndex );
	}

	return retval;
}

// 2019-08-24 ffgl@subpixels.com - ParamInfoStruct linked list replaced by params vector
FFResult PluginManager::SetParamElementValue( FFUInt32 paramIndex, FFUInt32 elemIndex, float value )
{
	FFResult retval = FF_FAIL;

	// Find the parameter (for editing)
	auto param = FindParam( paramIndex );
	if( !param )
		return retval;

	ParameterType paramType = param->GetType();

	if( paramType == ParameterType::FF_TYPE_BUFFER )
	{
		auto paramBuffer = dynamic_cast< ParamBuffer* >( param );
		if( paramBuffer && elemIndex < paramBuffer->data.size() )
		{
			// Update the buffer data element
			paramBuffer->data[ elemIndex ] = value;
			retval = FF_SUCCESS;
		}
	}

	return retval;
}

ParameterUsage PluginManager::GetParamUsage( FFUInt32 index ) const
{
	ParameterUsage retval = ParameterUsage::FF_FAIL;

	auto param = GetParam( index );
	if( !param )
		return retval;

	ParameterType paramType = param->GetType();

	if( paramType == ParameterType::FF_TYPE_BUFFER )
	{
		auto paramBuffer = dynamic_cast< const ParamBuffer* >( param );
		if( paramBuffer )
			retval = paramBuffer->GetUsage();
	}

	return retval;
}

// 2019-07-11 ffgl@subpixels.com - use specific FFMixed union members instead of pointer recasting
// 2019-07-22 ffgl@subpixels.com - use enum ParameterType
// 2019-08-24 ffgl@subpixels.com - ParamInfoStruct linked list replaced by params vector
FFMixed PluginManager::GetParamDefault( FFUInt32 index ) const
{
	FFMixed retval = { FF_FAIL };

	auto param = GetParam( index );
	if( !param )
		return retval;

	retval = param->GetMixedDefault();

	return retval;
}

FFResult PluginManager::ResetParam( FFUInt32 index )
{
	FFResult retval = FF_FAIL;

	auto param = FindParam( index );
	if( !param )
		return retval;

	param->Reset();
	return FF_SUCCESS;
}

bool PluginManager::GetTimeSupported() const
{
	return timeSupported;
}

// Find param (for editing); returns null if out of range
// 2019-08-22 ffgl@subpixels.com - created
ParamBase* PluginManager::FindParam( FFUInt32 index )
{
	return ( index < params.size() ) ? params[ index ] : nullptr;
}

// Get param (read only); returns null if out of range
// 2019-08-22 ffgl@subpixels.com - created
const ParamBase* PluginManager::GetParam( FFUInt32 index ) const
{
	return ( index < params.size() ) ? params[ index ] : nullptr;
}

}//End namespace ffgl
